/****************************************************************************
* Copyright (C) 2023 Matthias Hammelsbeck
*
*
* This file may be copied and/or distributed without explicit permit.
* Authors: Dominik Egert and Matthias Hammelsbeck
*****************************************************************************/

#include <Generic/HashMap.mqh>

/* The following defines coming from Dominik Egert (see https://www.mql5.com/en/forum/438987) */
#define new         dbg_heap_obj_trace.location_recorder(__LINE__, __FILE__) = new
#define delete      delete dbg_heap_obj_trace -


class Cdbg_heap_obj_trace
  {
private:
   // Local storage
   CHashMap<string,string>  m_map;
   int                      m_line;
   string                   m_filename;

public:
   // Constructor
     Cdbg_heap_obj_trace()
     {
      m_map.Clear();
     }

   // Destructor
     ~Cdbg_heap_obj_trace()
     {
      Print("-- Resume Memory Leaks --");
      string desc[], infos[];
      ArrayFree(desc);
      ArrayFree(infos);
      m_map.CopyTo(desc,infos);
      int size=ArraySize(infos);
      if(size == 0)
        {
         Print("  no memory leaks");
        }
      else
        {
         Print(StringFormat("%d undeleted objects left", size));
         for(int i=0; i<size; i++)
           {
            Print(StringFormat("  - %s", infos[i]));
           }
        }
      Print("-- End Resume --");
      m_map.Clear();
     }

   Cdbg_heap_obj_trace* location_recorder(const int line, const string filename)
     {
      m_line     = line;
      m_filename = filename;
      return(GetPointer(this));
     }

   template <typename V>
   V* operator=(V* p_in)
     {
      string info = StringFormat("File %s Line %d", m_filename, m_line);
      if(!m_map.Add(StringFormat("%x",p_in), info))
        {
         Print(StringFormat("%s ERROR m_map.Add", __FUNCTION__));
        }
      return p_in;
     }

   template <typename V>
   V* operator-(V* p_in)
     {
      string desc = StringFormat("%x",p_in);
      if(!m_map.ContainsKey(desc))
        {
         Print(StringFormat("ERROR at delete: Object %s not found!", desc));
         return p_in;
        }
      m_map.Remove(desc);
      return p_in;
     }
  } dbg_heap_obj_trace;
